//
// aniUI.cpp
//
// The user interface part for the program. 
//
#pragma warning(disable:4786)

#include <stdio.h>
#include <time.h>

#ifndef WIN32
#include <sys/times.h>
#endif

#include <fstream>

#include <FL/fl_ask.h>

#include "aniUI.h"
#include "aniDoc.h"

#include "glView.h"

#include "ctrlPanel.h"

#include "fltkCurveEditor.h"

#include "CurveEditor.h"

#include "Curves.h"

#include "ctrlPanel.h"
extern ctrlPanel* g_panel;
extern CAniDoc *aniDoc;
extern CAniUI *aniUI;

extern CurveEditor *curve_editor;

extern void CurveEditor_UpdateVariables();

int g_timerun;
float g_fps;
int g_mode = CURVE_MODE;
//bool g_useCurve;
bool g_viewCurve;
bool g_viewBezier;
float g_delay=-1;
int g_start=0, g_end=20;
float g_interval;

bool g_bSaveMovie=false;
int g_saveFrameCnt;
char g_saveTitle[128];

char *g_rayfilename = NULL;

static float g_currtime=0;
float getCurrentTime()
{	
	return g_currtime;
}

void setCurrentTime(float t)
{	
	g_currtime = t;
}
void timed_callback(void *)
{
	static clock_t last=0;

#ifdef WIN32
#define CLOCKS_PER_SECOND		CLOCKS_PER_SEC
	clock_t now=clock();
#else
#define CLOCKS_PER_SECOND			100
	struct tms ttms;
	clock_t now=times(&ttms);
#endif

	if (g_timerun!=0 && aniDoc->getScene()) {
		if (now!=last && last!=0)
			g_fps=0.9*g_fps+0.1*(CLOCKS_PER_SECOND/(double)(now-last));

		if (g_timerun<0 && last!=0) {
			g_currtime+=(double)(now-last)/CLOCKS_PER_SECOND;
		}
		else
			g_currtime+=1.0/g_timerun;


		if (g_currtime>=aniUI->m_pCurveEditor->getEndTime() 
			  || g_currtime<aniUI->m_pCurveEditor->getStartTime() ) {
			g_currtime=aniUI->m_pCurveEditor->getStartTime();
		} 

		aniUI->m_pCurveEditor->refresh();
		aniDoc->getScene()->refresh();

		Fl::flush();

		if (Fl::ready()) {
			// check event
			Fl::check();
		}

		if (g_fps<g_timerun) {	
			g_interval-=(1000.0f/g_fps-1000.0f/g_timerun)/2.0;		// g_interval in ms
		} else {
			g_interval+=(1000.0f/g_timerun-1000.0f/g_fps)/2.0;	;
		}

		if (g_interval<0 || g_timerun<0) g_interval=0;

		Fl::add_timeout(g_interval/1000.0f,timed_callback);
	} 

	last = now;
}

//------------------------------------- Help Functions --------------------------------------------

CAniUI* CAniUI::whoami(Fl_Menu_* o)	// from menu item back to UI itself
{
	return ( (CAniUI*)(o->parent()->user_data()) );
}

// utility function added so that aniMain can auto load a model if provided in the
// command line.  Code extracted from cb_load_object by DWJ 5/2000.

void CAniUI::load_named_object(CAniUI* pUI, CAniDoc* pDoc, char* libname)
{
	char buf[128];

	if (libname != NULL) {

		if (pDoc->loadObject(libname)) {
			sprintf(buf, "<%s>", libname);

			((Scene *)(pDoc->getScene()))->show();

		} else{
			//fl_alert("Can't load %s", libname);
			sprintf(buf, "Modeler/Animator <Not Loaded>");
		}

		pUI->m_mainWindow->label(buf);
	}
}

//--------------------------------- Callback Functions --------------------------------------------
void CAniUI::cb_load_object(Fl_Menu_* o, void* v)
{

	CAniUI* pUI=whoami(o);
	CAniDoc* pDoc=pUI->getDocument();

	if (g_fps!=0) {
		fl_alert("You need to turn off animation before loading a object!");
		return;
	}

#ifdef WIN32
	char* newfile = fl_file_chooser("Open Object?", "*.dll", NULL );
#else
	char* newfile = fl_file_chooser("Open Object?", "*.so", NULL );
#endif

	if (newfile != NULL) {
		// close the previous scene before we proceed
		pDoc->closeScene();
		pUI->closeAllViews();
		pUI->closeAllPanels();
		pUI->m_pCurveEditor->reset();
		// hide fltkCurveEditor
		pUI->m_pCurveEditor->hide();

		g_currtime=0;
		g_fps=0;

		pUI->load_named_object(pUI,pDoc,newfile);
	}
}

void CAniUI::cb_load_scene(Fl_Menu_* o, void* v)
{
	char buf[128];

	CAniUI* pUI=whoami(o);
	CAniDoc* pDoc=pUI->getDocument();

	if (g_fps!=0) {
		fl_alert("You need to turn off animation before loading a object!");
		return;
	}

	char* newfile = fl_file_chooser("Open Scene?", "*.scn", NULL );

	if (newfile != NULL) {
		// close the previous scene before we proceed
		pDoc->closeScene();
		pUI->closeAllViews();
		pUI->closeAllPanels();
		pUI->m_pCurveEditor->reset();
		// hide fltkCurveEditor
		pUI->m_pCurveEditor->hide();

		g_currtime=0;
		g_fps=0;

		if (pDoc->loadScene(newfile)) {
			sprintf(buf, "<%s>", newfile);

			((Scene *)(pDoc->getScene()))->show();

		} else{
			fl_alert("Can't load %s", newfile);
			sprintf(buf, "Modeler/Animator <Not Loaded>");
		}

		pUI->m_mainWindow->label(buf);
	}
}

void CAniUI::cb_close_scene(Fl_Menu_* o, void* v)
{
	CAniUI* pUI=whoami(o);
	CAniDoc* pDoc=pUI->getDocument();

	if (g_fps!=0) {
		fl_alert("You need to turn off animation before close object!");
		return;
	}

	pDoc->closeScene();
	pUI->closeAllViews();
	pUI->closeAllPanels();

	pUI->m_pCurveEditor->reset();
	
	// hide fltkCurveEditor
	pUI->m_pCurveEditor->hide();
}

void CAniUI::cb_load_animation(Fl_Menu_* o, void* v)
{
	char* newfile = fl_file_chooser("Load Animation?", "*.anm", NULL );

	if (newfile != NULL) {
		std::ifstream load_file(newfile);

		Curves::getInstance()->fromStream(load_file);
		
		load_file.close();

		CurveEditor_UpdateVariables();
	}
}

void CAniUI::load_named_ani(char* ani)
{
	if(ani)
	{
		std::ifstream load_file(ani);
		if(load_file)
		{
			Curves::getInstance()->fromStream(load_file);
			
			load_file.close();

			CurveEditor_UpdateVariables();
		}
	}
}



void CAniUI::cb_save_animation(Fl_Menu_* o, void* v)
{
	char* newfile = fl_file_chooser("Save Animation?", "*.anm", NULL );

	if (newfile != NULL) {
		std::ofstream save_file(newfile);

		Curves::getInstance()->toStream(save_file);
		
		save_file.close();

		CurveEditor_UpdateVariables();
	}
}

void CAniUI::cb_save_snapshot(Fl_Menu_* o, void* v)
{
	CAniUI* pUI=whoami(o);

	char* newfile = fl_file_chooser("Save Snapshot (bmp)?", "*.bmp", "save.bmp" );

	if (newfile != NULL) {
		pUI->m_pDoc->getScene()->saveImage(newfile);
	}
}

void CAniUI::cb_save_ray_snapshot(Fl_Menu_* o, void* v)
{
	CAniUI* pUI=whoami(o);

	g_rayfilename = fl_file_chooser("Save Snapshot (ray)?", "*.ray", "save.ray" );
	if (g_rayfilename != NULL) {

		// set draw mode for ray tracing

		int savemode = pUI->m_nDarwMode;
		pUI->m_nDarwMode=((pUI->m_nDarwMode&0xFFFFFF00)|MODE_RAYTRACE);

		// refresh the image, which will cause a dump of ray trace info, 
		// and wait for it to be displayed before restoring the mode.

		pUI->refreshAll();
		Fl::flush();
		if (Fl::ready())
		{
			Fl::check();
		}

		// restore draw mode and draw again for the screen

		pUI->m_nDarwMode = savemode;
		pUI->refreshAll();
		}
}

void CAniUI::cb_save_ray_movie(Fl_Menu_* o, void* v)
{
	CAniUI* pUI=whoami(o);
	static char buf[128];
	g_rayfilename = fl_file_chooser("Save Raytracer Movie?", "*.ray", "save" );
	if (g_rayfilename != NULL) {

		// set draw mode for ray tracing
		if (g_timerun==0) {
			g_timerun=30;
		}

		strcpy(buf, g_rayfilename);
		int savemode = pUI->m_nDarwMode;
		pUI->m_nDarwMode=((pUI->m_nDarwMode&0xFFFFFF00)|MODE_RAYTRACE);

		if (aniDoc->getScene()) {
			g_currtime=aniUI->m_pCurveEditor->getStartTime();

			while (g_currtime<aniUI->m_pCurveEditor->getEndTime()) {

				sprintf(g_rayfilename, "%s%05d.ray", buf, g_saveFrameCnt++);

				pUI->refreshAll();
				Fl::flush();
	
				if (Fl::ready()) {
					// check event
					Fl::check();
				}

				if (g_timerun<0) {
					g_currtime+=1/30.0f;	// 30 fps for actual time
				}
				else
					g_currtime+=1.0/g_timerun;
			}

			fl_message("Movie Done!");
		} 


		pUI->m_nDarwMode = savemode;
		pUI->refreshAll();
	}
}

void CAniUI::cb_save_movie(Fl_Menu_* o, void* v)
{
	CAniUI* pUI=whoami(o);

	if (g_timerun==0) {
		g_timerun=30;
	}

	char* newfile = fl_file_chooser("Save movie?", "*.bmp", "frame" );

	if (newfile != NULL) {
		g_bSaveMovie=true;
		g_saveFrameCnt=0;

		// parse title
		strcpy(g_saveTitle, newfile);		
		char *p=strrchr(g_saveTitle, '.');
		if (p!=NULL) {
			if (strcmp(p+1, "bmp")==0 || strcmp(p+1, "BMP")==0)
				*p=0;
		}

		if (aniDoc->getScene()) {
			g_currtime=aniUI->m_pCurveEditor->getStartTime();

			while (g_currtime<aniUI->m_pCurveEditor->getEndTime()) {

				aniUI->m_pCurveEditor->refresh();
				aniDoc->getScene()->refresh();
					
				Fl::flush();
	
				if (Fl::ready()) {
					// check event
					Fl::check();
				}

				static char buf[128];

				sprintf(buf, "%s%05d.bmp", g_saveTitle, g_saveFrameCnt++);

				aniDoc->getScene()->saveImage(buf);

				if (g_timerun<0) {
					g_currtime+=1/30.0f;	// 30 fps for actual time
				}
				else
					g_currtime+=1.0/g_timerun;
			}

			fl_message("Movie Done!");
		} 

		g_bSaveMovie=false;
	}
}

void CAniUI::cb_redraw(Fl_Menu_* o, void* v)
{
	CAniUI* pUI=whoami(o);

	pUI->refreshAll();
}

void CAniUI::cb_set_opmode(Fl_Menu_* o, void* v)
{
	CAniUI* pUI=whoami(o);
	CAniDoc* pDoc=pUI->getDocument();

	pUI->m_nOpMode=(int)v;

	if (((int)v)==CURVE_MODE) {
		g_mode=CURVE_MODE;

		// show fltkCurveEditor
		pUI->m_pCurveEditor->show();
		pUI->m_pCurveEditor->setScene(pUI->getDocument()->getScene());
	
		// hide the object panel
		if (pDoc->getScene()) {
			pDoc->getScene()->m_pActiveObj->m_pPanel->hide();
		}

		pUI->refreshAll();
	} else {
		g_mode=CONTROL_MODE;

		// hide fltkCurveEditor
		pUI->m_pCurveEditor->hide();

		// if the object doesn't currently have a subwindow, show the panel and create a scene for the object
		if (pDoc->getScene()) {
			pDoc->getScene()->m_pActiveObj->m_pPanel->show();
			pUI->addPanel(pDoc->getScene()->m_pActiveObj->m_pPanel);
		}

		pUI->refreshAll();
	}
}

void CAniUI::cb_set_drawmode(Fl_Menu_* o, void* v)
{
	CAniUI* pUI=whoami(o);

	pUI->m_nDarwMode=((pUI->m_nDarwMode&0xFFFFFF00)|(int)v);

	pUI->refreshAll();
}

void CAniUI::cb_set_drawquality(Fl_Menu_* o, void* v)
{
	CAniUI* pUI=whoami(o);

	pUI->m_nDarwMode=((pUI->m_nDarwMode&0xFFFF00FF)|(int)v);

	pUI->refreshAll();
}

void CAniUI::cb_exit(Fl_Menu_* o, void* v)
{
	CAniUI* pUI=whoami(o);
	CAniDoc* pDoc=pUI->getDocument();

	pUI->closeAllViews();
	pUI->closeAllPanels();
	pUI->m_mainWindow->hide();
	pUI->m_pCurveEditor->hide();
	if (pDoc->getScene() && pDoc->getScene()->m_pObjList)
		pDoc->getScene()->m_pObjList->hide();
	exit(0);
}

void CAniUI::cb_exit2(Fl_Widget* o, void* v) 
{
	CAniUI* pUI=(CAniUI*)(o->user_data());
	CAniDoc* pDoc=pUI->getDocument();
	
	pUI->closeAllViews();
	pUI->closeAllPanels();
	pUI->m_mainWindow->hide();
	pUI->m_pCurveEditor->hide();
	if (pDoc->getScene() && pDoc->getScene()->m_pObjList)
		pDoc->getScene()->m_pObjList->hide();
	exit(0);
}

void CAniUI::cb_fps(Fl_Menu_* o, void* v) 
{
	g_timerun = (int)v;
	g_fps = g_timerun;
	g_delay = -1;
	if (g_timerun<0) g_interval=0;
	else if (g_timerun!=0) g_interval = 1000.0f/g_timerun;
	timed_callback((void *)0);
}

void CAniUI::cb_reset_time(Fl_Menu_* o, void* v) 
{
	CAniUI* pUI=whoami(o);

	g_currtime=0;
	pUI->m_pCurveEditor->refresh();	
}


void CAniUI::cb_view_curve(Fl_Menu_* o, void* v)
{
	CAniUI* pUI=whoami(o);

	g_viewCurve=!g_viewCurve;
	pUI->refreshAll();
}
void CAniUI::cb_view_bezier(Fl_Menu_* o, void* v)
{
	CAniUI* pUI=whoami(o);

	g_viewBezier=!g_viewBezier;
	pUI->refreshAll();
}

void CAniUI::cb_about(Fl_Menu_* o, void* v) 
{
	fl_message("Modeler/Animator\nCSE457, Spring 2000");
//	fl_message("Modeler\nCSE457, Spring 2000");
}

//
// Member functions
//
void CAniUI::show() {
	m_mainWindow->show();
}

CAniDoc* CAniUI::getDocument()
{
	return m_pDoc;
}

void CAniUI::setDocument(CAniDoc *doc)
{
	m_pDoc = doc;
}

void CAniUI::refreshAll()
{
	typedef list<CGLView*>::const_iterator iter;

	for( iter j = glViews.begin(); j != glViews.end(); ++j ) {
		(*j)->refresh() ;
	}
}

void CAniUI::closeAllViews()
{
	typedef list<CGLView*>::const_iterator iter;

	for( iter j = glViews.begin(); j != glViews.end(); ++j ) {
		(*j)->hide() ;
	}
}

void CAniUI::closeAllPanels()
{
	typedef list<ctrlPanel*>::const_iterator iter;

	for( iter j = panels.begin(); j != panels.end(); ++j ) {
		(*j)->hide() ;
	}
}

// menu definition
Fl_Menu_Item CAniUI::menuitems[] = {
	{ "&File",		0, 0, 0, FL_SUBMENU },
		{ "&Load Object...",	FL_ALT + 'l', (Fl_Callback *)CAniUI::cb_load_object },
		{ "&Close Obejct...",	FL_ALT + 'c', (Fl_Callback *)CAniUI::cb_close_scene, 0, FL_MENU_DIVIDER },
		{ "&Load Scene...",	FL_ALT + 'l', (Fl_Callback *)CAniUI::cb_load_scene },
		{ "&Close Scene...",	FL_ALT + 'c', (Fl_Callback *)CAniUI::cb_close_scene, 0, FL_MENU_DIVIDER },
		{ "&Load Animation...",	0, (Fl_Callback *)CAniUI::cb_load_animation },
		{ "&Save Animation...",	0, (Fl_Callback *)CAniUI::cb_save_animation, 0, FL_MENU_DIVIDER },
		{ "&Save BMP Snapshot ...",	0, (Fl_Callback *)CAniUI::cb_save_snapshot },
		{ "&Save RAY Snapshot ...",	0, (Fl_Callback *)CAniUI::cb_save_ray_snapshot, 0, FL_MENU_DIVIDER},
		{ "&Save BMP Movie...",	0, (Fl_Callback *)CAniUI::cb_save_movie },
		{ "&Save RAY Movie...",	0, (Fl_Callback *)CAniUI::cb_save_ray_movie, 0, FL_MENU_DIVIDER },
		{ "&Exit",			FL_ALT + 'e', (Fl_Callback *)CAniUI::cb_exit },
		{ 0 },

	{ "&Edit",		0, 0, 0, FL_SUBMENU },
		{ "&Redraw",		FL_ALT + 'r', (Fl_Callback *)CAniUI::cb_redraw, 0, FL_MENU_DIVIDER },
		{ "Con&trol Mode",	FL_ALT + 't', (Fl_Callback *)CAniUI::cb_set_opmode, (void *)CONTROL_MODE, FL_MENU_RADIO  },
		{ "&Curve Mode",	FL_ALT + 'c', (Fl_Callback *)CAniUI::cb_set_opmode, (void *)CURVE_MODE, FL_MENU_RADIO|FL_MENU_DIVIDER |FL_MENU_VALUE},
		{ 0 },

	{ "&View",		0, 0, 0, FL_SUBMENU },
		{ "&Normal",		FL_ALT + 'n', (Fl_Callback *)CAniUI::cb_set_drawmode, (void *)MODE_NORMAL, FL_MENU_RADIO|FL_MENU_VALUE },
		{ "&Wireframe",		FL_ALT + 'w', (Fl_Callback *)CAniUI::cb_set_drawmode, (void *)MODE_WIREFRAME, FL_MENU_RADIO },
		{ "&Flat Shaded",	FL_ALT + 'f', (Fl_Callback *)CAniUI::cb_set_drawmode, (void *)MODE_FLATSHADE, FL_MENU_RADIO|FL_MENU_DIVIDER },
		{ "&High Quality",	FL_ALT + 'h', (Fl_Callback *)CAniUI::cb_set_drawquality, (void *)MODE_HIGH_QUALITY, FL_MENU_RADIO },
		{ "&Medium Quality",FL_ALT + 'm', (Fl_Callback *)CAniUI::cb_set_drawquality, (void *)MODE_MED_QUALITY, FL_MENU_RADIO },
		{ "&Low Quality",	FL_ALT + 'l', (Fl_Callback *)CAniUI::cb_set_drawquality, (void *)MODE_LOW_QUALITY, FL_MENU_RADIO|FL_MENU_VALUE },
		{ "&Poor Quality",	FL_ALT + 'p', (Fl_Callback *)CAniUI::cb_set_drawquality, (void *)MODE_POOR_QUALITY, FL_MENU_RADIO|FL_MENU_DIVIDER },
		{ "View &Curve Points", FL_ALT + 'c', (Fl_Callback *)CAniUI::cb_view_curve, 0, FL_MENU_TOGGLE},
		{ "View &Bezier Points", FL_ALT + 'b', (Fl_Callback *)CAniUI::cb_view_bezier, 0, FL_MENU_TOGGLE},

		{ 0 },

	{ "&Animate",		0, 0, 0, FL_SUBMENU },
		{ "&Animation Off",	0, (Fl_Callback *)CAniUI::cb_fps, (void *)0, FL_MENU_RADIO|FL_MENU_VALUE },
		{ "&5 frames/second", 0, (Fl_Callback *)CAniUI::cb_fps, (void *)5, FL_MENU_RADIO },
		{ "&10 frames/second", 0, (Fl_Callback *)CAniUI::cb_fps, (void *)10, FL_MENU_RADIO },
		{ "&20 frames/second", 0, (Fl_Callback *)CAniUI::cb_fps, (void *)20, FL_MENU_RADIO},
		{ "&30 frames/second", 0, (Fl_Callback *)CAniUI::cb_fps, (void *)30, FL_MENU_RADIO },
		{ "&Actual Time", 0, (Fl_Callback *)CAniUI::cb_fps, (void *)-1, FL_MENU_DIVIDER|FL_MENU_RADIO },
		{ "&Reset time", 0, (Fl_Callback *)CAniUI::cb_reset_time},
		{ 0 },

	{ "&Help", 0,  0, 0, FL_SUBMENU},
		{ "&About", 0, (Fl_Callback *)CAniUI::cb_about},
		{ 0 },

	{ 0 }
};

CAniUI::CAniUI() : glViews() {
	// init. values
	m_nDarwMode = MODE_NORMAL|MODE_LOW_QUALITY;

	m_mainWindow = new Fl_Window(100, 40, 320, 25, "Modeler/Animator (CSE 457)");
	//m_mainWindow = new Fl_Window(100, 40, 320, 25, "Modeler (CSE 457)");
		m_mainWindow->user_data((void*)(this));	// record self to be used by static callback functions
		// install menu bar
		m_menubar = new Fl_Menu_Bar(0, 0, 320, 25);
		m_menubar->menu(menuitems);

		m_mainWindow->callback(cb_exit2);
		m_mainWindow->when(FL_HIDE);
    m_mainWindow->end();

	m_pCurveEditor=new fltkCurveEditor(100, 100, 450, 600, "Curve Editor"); 
}

CAniUI::~CAniUI() 
{
	delete m_pCurveEditor;
}